structure DeflateOutput :> DeflateOutput =
struct

	type 'a container = (int * 'a list list list list list 
					* 'a list list list list
					* 'a list list list 
					* 'a list list 
					* 'a list)
	(* 
	   This is a multiple level skip list,
	   defined by (position, skipN ... skip0)
	   Each skipN is a list of skipN-1 lists, that is 
	   really pointers within skipN-1.
	   
	   A simple example would look like below
	   (^ means 'point to the list node above')
	   
	   skip0 = 1 :: 2 :: 3 :: 4 :: 5::nil   skip0_size=1  pos=5
	   skip1 =      ^ ::      ^ ::    nil   skip1_size=2
	   skip2 =      ^ ::              nil   skip2_size=4
       
	   It is not defined recursively with skipN..skip0 a list 
	   because this would make a polymorphic list which is not 
	   type-able (I think).
	 *)
		
	(*  skip0_size = 1*)
	val skip1_size = 8
	val skip2_size = skip1_size * skip1_size
	val skip3_size = skip2_size * skip1_size
	val skip4_size = skip3_size * skip1_size (* 4096 *)
	val full_size = DeflateBits.pow2 15
	
	val empty = (0, [], [], [], [], [])
	
	fun output (_, _, _, _, _, list) = rev list
	
	fun to_string (_, skip4, skip3, skip2, skip1, skip0) = "("^
		Int.toString (length skip4) ^","^
		Int.toString (length skip3) ^","^
		Int.toString (length skip2) ^","^
		Int.toString (length skip1) ^")"
		
	fun trim (cont as (pos, skip4, skip3, skip2, skip1, skip0)) =
		let
			fun f skips skipsize = DeflateList.trim skips (1 + Int.quot (full_size, skipsize))
		in
			if Int.rem (pos, 2*full_size) <> 0 then
				cont
			else
				(pos,
				f skip4 skip4_size, f skip3 skip3_size, 
				f skip2 skip2_size, f skip1 skip1_size,
				skip0)
		end

	fun append_byte (pos, skip4, skip3, skip2, skip1, skip0) v =
		let
			val p = pos + 1
			val new0 = v::skip0
		in
			if Int.rem (p, skip1_size) <> 0 then 
				(p, skip4, skip3, skip2, skip1, new0)
			else 
				let	val new1 = new0::skip1 
				in if Int.rem (p, skip2_size) <> 0 then
					(p, skip4, skip3, skip2, new1, new0)
				else
					let val new2 = new1::skip2 
					in if Int.rem (p, skip3_size) <> 0 then
						(p, skip4, skip3, new2, new1, new0)
					else
						let val new3 = new2::skip3 
						in if Int.rem (p,skip4_size) <> 0 then
							(p, skip4, new3, new2, new1, new0)
						else
							trim (p, new3::skip4, new3, new2, new1, new0)
						end
					end
				end
		end
		
		
	fun drop (pos, skip4, skip3, skip2, skip1, skip0) nbytes =
		let
			(* drop and go down, once aligned going down preserve alignment *)
			val quot = Int.quot
			val rem = Int.rem
			fun drop0 skips n = List.drop (skips, n)
			fun drop1 skips n = drop0 (hd (List.drop (skips, quot (n, skip1_size)))) (rem (n, skip1_size))
			fun drop2 skips n = drop1 (hd (List.drop (skips, quot (n, skip2_size)))) (rem (n, skip2_size))
			fun drop3 skips n = drop2 (hd (List.drop (skips, quot (n, skip3_size)))) (rem (n, skip3_size))
			fun drop4 skips n = drop3 (hd (List.drop (skips, quot (n, skip4_size)))) (rem (n, skip4_size))
		in
			(* find first level to be aligned and then skip down*)
			if nbytes > rem (pos, skip4_size) then
				drop4 skip4 (nbytes - rem (pos, skip4_size))
			else if nbytes > rem (pos, skip3_size) then
				drop3 skip3 (nbytes - rem (pos, skip3_size))
			else if nbytes > rem (pos, skip2_size) then
				drop2 skip2 (nbytes - rem (pos, skip2_size))
			else if nbytes > rem (pos, skip1_size) then
				drop1 skip1 (nbytes - rem (pos, skip1_size))
			else
				List.drop (skip0, nbytes)
		end
			
	fun append_bytes output [] = output
	  | append_bytes output (b::bytes) = 
			append_bytes (append_byte output b) bytes

	fun append_from_output output 0 _ = output
	  | append_from_output output nbytes distance =
	  		if distance >= nbytes then
				append_bytes
					output 
					(rev (List.take (drop output (distance - nbytes), nbytes)))
			else
				(* split overlapping copy *)
				append_from_output 
					(append_from_output output distance distance)
					(nbytes - distance)
					distance
	
	(* self test *)
	val tests = [
		output (append_from_output (append_byte empty 111) 1 1) = [111,111]
		]
end
